"""
This sample illustrates how to get images from the blaze camera using the Python Harvester module.
"""

import os
import platform

## This is used for reshaping the image buffers.
import numpy as np

## This is used for visualization.
import cv2

## Use of Harvester to access the camera.
## For more information regarding Harvester, visit the github page:
## https://github.com/genicam/harvesters
from harvesters.core import Harvester

h = Harvester()

## Location of the Basler blaze GenTL producer.
if platform.system() == "Windows":
    path_to_blaze_cti = os.environ['BASLER_BLAZE_BIN64'] + "\\ProducerBlaze.cti"
elif platform.system() == "Linux":
    path_to_blaze_cti = "/opt/Basler_blaze/lib/gentlproducer/gtl/ProducerBlaze.cti"
else:
    print(f"{platform.system()} is not supported")
    assert False

## Add producer to Harvester.
assert os.path.exists(path_to_blaze_cti)
h.add_file(path_to_blaze_cti)

## Update device list.
h.update()

## Print device list.
print(h.device_info_list)

## Connect to the first camera in the device list.
if h.device_info_list:
    ia = h.create_image_acquirer(0)
else:
    print("No cameras found")
    exit()

## In the following, we demonstrate how to get and set camera parameters.
## For demonstration purposes and to avoid having to change the camera's state,
## we first get a parameter value and then set it again.

## Access OperatingMode.
operatingMode = ia.remote_device.node_map.OperatingMode.value
ia.remote_device.node_map.OperatingMode.value = operatingMode
# ia.remote_device.node_map.OperatingMode.value = 'ShortRange'
print("OperatingMode: ", ia.remote_device.node_map.OperatingMode.value)

## Access FastMode.
fastMode = ia.remote_device.node_map.FastMode.value
ia.remote_device.node_map.FastMode.value = fastMode
# ia.remote_device.node_map.FastMode.value = True
print("FastMode: ", ia.remote_device.node_map.FastMode.value)

## Access FilterSpatial
filterSpatial = ia.remote_device.node_map.FilterSpatial.value
ia.remote_device.node_map.FilterSpatial.value = filterSpatial
# ia.remote_device.node_map.FilterSpatial.value = True
print("FilterSpatial: ", ia.remote_device.node_map.FilterSpatial.value)

## Access FilterTemporal
filterTemporal = ia.remote_device.node_map.FilterTemporal.value
ia.remote_device.node_map.FilterTemporal.value = filterTemporal
# ia.remote_device.node_map.FilterTemporal.value = True
print("FilterTemporal: ", ia.remote_device.node_map.FilterTemporal.value)

## Access FilterStrength
filterStrength = ia.remote_device.node_map.FilterStrength.value
ia.remote_device.node_map.FilterStrength.value = filterStrength
## FilterTemporal must be enabled before setting
## FilterStrengh is possible.
if ia.remote_device.node_map.FilterTemporal.value:
    ia.remote_device.node_map.FilterStrength.value = filterStrength
    # ia.remote_device.node_map.FilterStrength.value = 200
print("FilterStrength: ", ia.remote_device.node_map.FilterStrength.value)

## Access OutlierRemoval
outlierRemoval = ia.remote_device.node_map.OutlierRemoval.value
ia.remote_device.node_map.OutlierRemoval.value = outlierRemoval
# ia.remote_device.node_map.OutlierRemoval.value = True
print("OutlierRemoval: ", ia.remote_device.node_map.OutlierRemoval.value)

## Access ConfidenceThreshold
confidenceThreshold = ia.remote_device.node_map.ConfidenceThreshold.value
ia.remote_device.node_map.ConfidenceThreshold.value = confidenceThreshold
# ia.remote_device.node_map.ConfidenceThreshold.value = 20
print("ConfidenceThreshold: ", ia.remote_device.node_map.ConfidenceThreshold.value)

## Access GammaCorrection
gammaCorrection = ia.remote_device.node_map.GammaCorrection.value
ia.remote_device.node_map.GammaCorrection.value = gammaCorrection
# ia.remote_device.node_map.GammaCorrection.value = True
print("GammaCorrection: ", ia.remote_device.node_map.GammaCorrection.value)

## Set the working range to the values displayed for the Max. Depth [mm] and
## Min. Depth [mm] parameters.
## The working range depends on the current operating mode, so
## the OperatingMode parameter must be set before adjusting DepthMax and DepthMin.
## LongRange: 0 .. 9993mm
## ShortRange: 0 .. 1498mm
depthMin = ia.remote_device.node_map.DepthMin.value
depthMax = ia.remote_device.node_map.DepthMax.value
ia.remote_device.node_map.DepthMin.value = depthMin
ia.remote_device.node_map.DepthMax.value = depthMax
# ia.remote_device.node_map.DepthMin.value = 0
# ia.remote_device.node_map.DepthMax.value = 9993
print("Min. Depth [mm]: ", ia.remote_device.node_map.DepthMin.value)
print("Max. Depth [mm]: ", ia.remote_device.node_map.DepthMax.value)


## Control pixel formats for image components.
## Range information can be sent either as a 16-bit grey value image or as
## 3D coordinates (point cloud).
## For this sample, we want to acquire 3D coordinates.
## Note: To change the format of an image component, the Component Selector parameter
## must first be set to the component
## you want to configure.
## To use 16-bit integer depth information, choose "Coord3D_C16" instead of "Coord3D_ABC32f".
ia.remote_device.node_map.ComponentSelector.value = "Range"
ia.remote_device.node_map.ComponentEnable.value = True
ia.remote_device.node_map.PixelFormat.value = "Coord3D_ABC32f"

ia.remote_device.node_map.ComponentSelector.value = "Intensity"
ia.remote_device.node_map.ComponentEnable.value = True
ia.remote_device.node_map.PixelFormat.value = "Mono16"

ia.remote_device.node_map.ComponentSelector.value = "Confidence"
ia.remote_device.node_map.ComponentEnable.value = True
ia.remote_device.node_map.PixelFormat.value = "Confidence8"

print('To exit, press ESC in one of the image windows')
## Start image acquisition.
ia.start_acquisition()

while True:
    with ia.fetch_buffer() as buffer:
        ## Warning: The buffer is only valid in the with statement and will be destroyed
        ## when you leave the scope.
        ## If you want to use the buffers outside of the with scope, you have to use np.copy()
        ## to make a deep copy of the image.

        ## Create an alias of the image components:
        pointcloud = buffer.payload.components[0]

        intensity = buffer.payload.components[1]

        confidence = buffer.payload.components[2]

        ## Reshape the range image into a 2D/3D array:
        ## "num_components_per_pixel" depends on the pixel format selected:
        ## "Coord3D_ABC32f" = 3
        ## "Coord3D_C16" = 1
        _3d = pointcloud.data.reshape(pointcloud.height, pointcloud.width, \
                                        int(pointcloud.num_components_per_pixel))

        ## Reshape the intensity image into a 2D array:
        _2d_intensity = intensity.data.reshape(intensity.height, intensity.width)

        ## Reshape the confidence image into a 2D array:
        _2d_confidence = confidence.data.reshape(confidence.height, confidence.width)

        ## Show the captured images as grayscale.
        ## We only show the z-component of the point cloud.
        ## If you choose "Coord3D_C16" as pixel format, you have to remove [:,:,2].
        ## OpenCV can't show float values. We convert it for visualization to uint8.
        _3d_scaled = _3d * 255.0 / ia.remote_device.node_map.DepthMax.value
        cv2.imshow('range', _3d_scaled[:, :, 2].astype(np.uint8))

        cv2.imshow('_2d_intensity', _2d_intensity)

        cv2.imshow('_2d_confidence', _2d_confidence)

    ## Break the endless loop by pressing ESC.
    k = cv2.waitKey(5) & 0xFF
    if k == 27:
        break

## Stop image acquisition.
ia.stop_acquisition()

## Disconnect from camera.
ia.destroy()

## Remove the CTI file and reset Harvester.
h.reset()
